#!/usr/bin/perl

use strict;
use warnings;
use Getopt::Long;
use IO::Socket;
use Fcntl qw(:flock);
$|++;

# Overview: Downloads files from a variety of websites, measures the
# following characeristics 1. Throughput 2. Latency 3. Traceroute
#
# @author: Katrina LaCurts

# Struct for each site - Keep track of site address, the file we
# downloaded, its size, the time it took us to download it, the
# traceroute output to that site, and the output from wget (from
# getting the file; this might not be used in our analysis)
use Class::Struct site_info => {file => '$', dl_file_size => '$', dl_time => '$', traceroute => '$', wget_latency => '$', ping_latency => '$', ul_time_mit => '$', ul_file_size_mit => '$', ul_external_time => '$', ul_external_file_size => '$', ul_external_location => '$', ul_external_latency => '$'};

##################### Prototypes

sub initialize();
sub get_meta_data();
sub get_sites();
sub download_files();
sub send_output();
sub do_traceroute(@);
sub do_download(@);
sub do_latency(@);

##################### Global Variables

# Array that stores the site information for each site
my @sites;
# Time at which the script was started
my $start_time = time();
# Are we doing traceroute?
my $traceroute_on = 0;
# Debugging
my $debug_output = 0;
# Input file with list of websites
my $dirname = `dirname $0`; chomp($dirname);
my $website_file = $dirname . "/websites";
# Output directory
my $output_dir = "output_dir";
# Commandline variables for testing
my $testing;
my $no_output;
my $no_upload;
my $override;

# my id
my $id_file = $dirname . "/my_id";
my $my_id = `tail -n 1 $id_file`;
chomp($my_id) if ($my_id);
$my_id = -1 if (!$my_id);

# Metadata for user
my $isp;
my $ip;
my $os;

# Metadata for script
my $version = 2;

##################### Function Calls

# Before this file is ever used, the machines have done two things:
#   (i) Synced with an NTP server
#   (ii) Set up a cron job to run this script every n minutes

# -1. Initialize.  Get commandline options
initialize();
# 0. Get some meta-data (users's IP)
get_meta_data();
# 1. Read in the websites file
get_sites();
# 2. Download the files from the websites, record the appropriate
# statistics (basically, do everything)
download_files();
# 3. Send the output to the server
send_output();

##################### Primary Functions

# Initialize.  Get some options.
sub initialize() {

    # The testing option isn't really a 'public' option, so
    # we don't print it in the help output
    GetOptions("testing!"=>\$testing,
			   "traceroute!"=>\$traceroute_on,
			   "debugging!"=>\$debug_output,
			   "input=s"=>\$website_file,
			   "output=s"=>\$output_dir,
			   "override!"=>\$override,
			   "nooutput!"=>\$no_output,
			   "noupload!"=>\$no_upload);

	# Only want one instance of this running at once
	my $lockfile="lockfile";
	`touch $lockfile` if (! -f $lockfile);
	open IN, $lockfile;
	if (!(flock(IN, LOCK_EX|LOCK_NB))) {
		print "Exiting.. instance already running\n" if ($debug_output);
		exit(1);
	}
}

# Gets some meta-data about this user (right now, just the IP)
sub get_meta_data() {

    # Get the user's OS
    $os = `uname`; chomp($os);

	# We used to try to get the ip and isp.  Don't bother with this
	# anymore.  The server can get the ip when we send it the data,
	# and we have an out-of-band method for getting the isp

#    $ip= `wget http://www.whatismyip.com/automation/n09230945.asp -O - -o /dev/null`;
#    $isp=`dig -x $ip 2>&1 | grep -A 1 "AUTHORITY SECTION" | grep -v "AUTHORITY SECTION" | awk '{print \$5}'`;
}

# Reads in the website file
sub get_sites() {

    open FILE, $website_file;
    while (<FILE>) {
		my @line = $_;
		chomp($line[-1]);
	
		# We allow comments in the websites file; ignore those
		next if (substr($line[0], 0, 1) eq "#");
	
		# Get address and file name, add to sites array
		my $file = $line[0];
	
		my $site_info = new site_info;
		$site_info->file($file);
	
		push(@sites, $site_info);
    }
    close FILE;
}

# Calls the functions that do the real work, stores the output
sub download_files() {

	my $done_ul = 0;

    foreach my $site_info (@sites) {
	
		my $file = $site_info->file;
		print "On file " . $file . "\n" if ($debug_output);
		
		# Get the traceroute output to this site, store it
		if (!$testing || $override) {
			print "Doing traceroute..\n" if ($debug_output);
			$site_info->traceroute(do_traceroute($file));
		}
		
		# Get the download time and output (generally output from wget), store it
		if (!$testing) {
			print "Downloading..\n" if ($debug_output);
			my ($dl_time, $dl_file_size) = do_download($file);
			$site_info->dl_time($dl_time);
			$site_info->dl_file_size($dl_file_size);
		}
		
		# Same thing, but now for latency
		if (!$testing) {
			print "Doing Latency..\n" if ($debug_output);
			my ($wget_latency, $ping_latency) = do_latency($file);
			$site_info->wget_latency($wget_latency);
			$site_info->ping_latency($ping_latency);
			print "latency: " . $site_info->wget_latency . " " . $site_info->ping_latency . "\n" if ($debug_output);
		}
		
		# Get upload time, store it
		if ((!$testing || $override) && $done_ul == 0) {
			$done_ul = 1;
			print "Doing Upload..\n" if ($debug_output);
			my ($ul_time, $ul_size) = do_upload_mit();
			print "upload time: " . $ul_time . "\n" if ($debug_output);
			$site_info->ul_time_mit($ul_time);
			$site_info->ul_file_size_mit($ul_size);
			my ($ul_external_time, $ul_external_size, $ul_external_latency, $ul_external_location) = do_upload_external();
			print "external upload time and latency: " . $ul_external_time . " " . $ul_external_latency . "\n" if ($debug_output);
			$site_info->ul_external_time($ul_external_time);
			$site_info->ul_external_file_size($ul_external_size);
			$site_info->ul_external_latency($ul_external_latency);
			$site_info->ul_external_location($ul_external_location);
		}
    }
}

# Sends the output to a server (running on cilantro)
sub send_output() {

	return if ($no_output);

    # Socket to cilantro.csail.mit.edu on port 7070
    my $sock = new IO::Socket::INET ( PeerAddr => '128.30.76.56',
									  PeerPort => '7070',
									  Proto => 'tcp');
    die "Failed to connect to server" if (!$sock);

	$ip="UNKNOWN" if (!$ip);
	$isp="UNKNOWN" if (!$isp);
	$os="UNKNOWN" if (!$os);

    print $sock "ip: " . $ip . "\n";
    print $sock "os: " . $os . "\n";
    print $sock "start_time: " . $start_time . "\n";
	print $sock "my_id: " . $my_id . "\n";
	print $sock "isp: " . $isp . "\n";
	print $sock "version: " . $version . "\n";

    foreach my $site_info (@sites) {
		print_site_info_db($site_info, $sock);
    }
    close($sock);
}

##################### Helper Functions

# Returns the output of a traceroute to the address given.  Assumes
# that the address is given as protocol://url
sub do_traceroute(@) {

	print "traceroute is on..\n" if ($traceroute_on);

    return if (!$traceroute_on);
	# We only want to do traceroutes once per day
	my @time_array = gmtime(int(time()));
	# Gives us the hour
	print "time_array[2] is " . $time_array[2] . "\n";
	return if ($time_array[2] != 18 && !$override); # Would prefer to use 7, not 16; just testing for now
	print "doing traceroute for real\n";

    my @split_address = split(/\//, $_[0]); # separate protocol from url
    my $address = $split_address[2]; # 2 not 1 because we have protocol//url.  we split on / to remove the / from .com/, e.g.
    
    # -w and -q to make traceroute a little snappier
	my $tr_result = `traceroute -w 2 -q 1 $address 2>&1`;
    
    chomp($tr_result) if ($tr_result);
    return $tr_result;
}

sub do_upload_external() {

	return if ($no_upload);

	my $location = "128.8.126.25";
    
	# Upload the data
    my $sock = new IO::Socket::INET ( PeerAddr => $location,
									  PeerPort => '7071',
									  Proto => 'tcp');
	print "couldn't upload\n" if (!$sock && $debug_output);
    return (0, 0) if (!$sock);

	# Set up the string
    my $file_size = 8000000; # in bytes
	if ($my_id == "8860" || $my_id == "9154") {
		$file_size = 500000; # special, for one of the boxes
	}
	my $s = "";
    for (my $b = 0; $b < $file_size; $b++) {
		$s = $s . "k";
    }

	my ($start_sec, $start_msec) = gettime();
	print $sock $s;
	my ($end_sec, $end_msec) = gettime();

    close($sock);

	# Get latency
	my $latency = `ping -c 1 $location | tail -n 1 | awk '{print \$4}' | awk 'BEGIN {FS=\"/\"} {print \$2}'`;
	chomp($latency) if ($latency);

    # Return the time that it took to DL as well as the output from wget and the file size
	return time_difference($start_sec, $start_msec, $end_sec, $end_msec), $file_size, $latency, $location;
}


sub do_upload_mit() {

	return if ($no_upload);
    
    my $sock = new IO::Socket::INET ( PeerAddr => '128.30.76.56',
									  PeerPort => '7071',
									  Proto => 'tcp');
	print "couldn't upload\n" if (!$sock && $debug_output);
	return (0, 0) if (!$sock);

	# Set up the string
    my $file_size = 8000000; # in bytes (8MB)
	if ($my_id == "8860" || $my_id == "9154") {
		$file_size = 500000; # special, for one of the boxes
	}
	my $s = "";
    for (my $b = 0; $b < $file_size; $b++) {
		$s = $s . "k";
    }

	my ($start_sec, $start_msec) = gettime();
	print $sock $s;
	my ($end_sec, $end_msec) = gettime();

    close($sock);
    
    # Return the time that it took to DL as well as the output from wget and the file size
    return time_difference($start_sec, $start_msec, $end_sec, $end_msec), $file_size;
}


# Downloads a given file from a given address, returns download time
# and download output (usually output from wget)
sub do_download(@) {
    
    my $file = $_[0];
    
    # Just wget the file to download it (could use curl, but merakis don't have it)
    my $download_command = "wget $file -O /dev/null";

	my ($start_sec, $start_msec) = gettime();
    my $output=`$download_command 2>&1`;
	my ($end_sec, $end_msec) = gettime();

    chomp($output) if ($output);
    chomp($output) if ($output);
    my $dl_file_size = 0;
    
    if ($output =~ m/saved \[([0-9]*)\/[0-9]*\]/) {
		$dl_file_size = $1;
    }
    
    # Return the time that it took to DL as well as the output from wget and the file size
    return time_difference($start_sec, $start_msec, $end_sec, $end_msec), $dl_file_size;
}

# Sends a wget --spider request to a given file at a given address.
# Returns the time the request took (the latency) and the output
sub do_latency(@) {
    
    my $file = $_[0];
    my @split_address = split(/\//, $file); # separate protocol from url
    my $location = $split_address[2]; # 2 not 1 because we have protocol//url.  we split on / to remove the / from .com/, e.g.

	# Ping latency
	my $ping_latency;
	$ping_latency = `ping -c 1 $location | tail -n 1 | awk '{print \$4}' | awk 'BEGIN {FS=\"/\"} {print \$2}'` if (!($location =~ m/hivelocity/));
	chomp($ping_latency) if ($ping_latency);
    
    # wget latency
    my $download_command = "wget --spider $file";

    my $output;
	my $num_times = 10;
	# Have to do it multiple times because our perl module to get the time sucks on the soekris boxes
    my ($start_sec, $start_msec) = gettime();
	for (my $i=0; $i<$num_times; $i++) {
		$output=`$download_command 2>&1`;
	}
    my ($end_sec, $end_msec) = gettime();
    chomp($output) if ($output); chomp($output) if ($output);

    # Return the two latencies
	my $wget_latency = time_difference($start_sec, $start_msec, $end_sec, $end_msec)/$num_times;
	return $wget_latency, $ping_latency;
}

# Returns time diff in milliseconds, given four arguments:
# start_time_seconds, start_time_microseconds
# end_time_seconds, end_time_microseconds
# (convenient if you use the array form of the output from gettimeofday)
sub time_difference(@) {
    
    # Nothing fancy; convert to regular numbers and subtract
    my ($start_sec, $start_msec, $end_sec, $end_msec) = @_;
    my $start_time = $start_sec + $start_msec/(10**6);
    my $end_time = $end_sec + $end_msec/(10**6);
    my $time_diff = ($end_time - $start_time)*10**3;
    
    return $time_diff;
}

########### Helper functions

sub gettime() {

	return (time(), 0);

    eval "use Time::HiRes";
    if (!$@) {
		return Time::HiRes::gettimeofday();
    } else {
		return (time(), 0);
    }
}

# Prints the info for one site (specifically for DB use)
sub print_site_info_db(@) {

    my $info = $_[0];
    my $out_file = $_[1];

    $out_file = \*STDOUT if (!$out_file);

    my $file = $info->file;
    my $dl_file_size = $info->dl_file_size;
    my $dl_time = $info->dl_time;
    my $wget_latency = $info->wget_latency;
	my $ping_latency = $info->ping_latency;
    my $ul_time = $info->ul_time_mit;
    my $ul_file_size = $info->ul_file_size_mit;
    my $traceroute = $info->traceroute;
	my $ul_external_time = $info->ul_external_time;
	my $ul_external_file_size = $info->ul_external_file_size;
	my $ul_external_latency = $info->ul_external_latency;
	my $ul_external_location = $info->ul_external_location;

    $file = "NULL" if (!$file);
    $dl_file_size = "NULL" if (!$dl_file_size);
    $dl_time = "NULL" if (!$dl_time);

    if (!$ul_time && !$ul_file_size) {
		$ul_time = "NULL";
	} elsif (!$ul_time) {
		$ul_time = 0;
	}
    $ul_file_size = "NULL" if (!$ul_file_size);
    $wget_latency = "NULL" if (!$wget_latency);
	$ping_latency = "NULL" if (!$ping_latency);
    $traceroute = "NULL" if (!$traceroute);
	$ul_external_time = "NULL" if (!$ul_external_time);
	$ul_external_file_size = "NULL" if (!$ul_external_file_size);
	$ul_external_latency = "NULL" if (!$ul_external_latency);
	$ul_external_location = "NULL" if (!$ul_external_location);

	my $out_string = "file: " . $file . " " .
	$dl_file_size . " " . $dl_time . " " .
	$ul_file_size . " " . $ul_time . " " .
	$ul_external_file_size . " " . $ul_external_time . " " .
	$ul_external_latency . " " . $ul_external_location . " " .
	$wget_latency . " " . " " . $ping_latency . " " .
	$traceroute . " TR_DONE\n";

    print $out_file $out_string;
}
